// use super::plotter::Plotter;
use axum::http::header;
use axum::response::IntoResponse;
use image::ImageFormat;
use plotters::prelude::*;
use std::io::{BufWriter, Cursor};

#[allow(clippy::upper_case_acronyms)]
pub enum BitMapOutputFormat {
    PNG,
}

pub struct BitMapPlotter {}
pub type BitMapPlotRoot<'a> =
    plotters::drawing::DrawingArea<plotters::backend::BitMapBackend<'a>, plotters::coord::Shift>;
pub type BitMapPlotFun = Box<dyn Fn(BitMapPlotRoot)>;

impl BitMapPlotter {
    pub fn examplefun() -> BitMapPlotFun {
        Box::new(|root: BitMapPlotRoot| {
            root.fill(&WHITE).unwrap();
            let mut chart = ChartBuilder::on(&root)
                .caption("y=x^2", ("sans-serif", 50).into_font())
                .margin(5)
                .x_label_area_size(30)
                .y_label_area_size(30)
                .build_cartesian_2d(-1f32..1f32, -0.1f32..1f32)
                .unwrap();

            chart.configure_mesh().draw().unwrap();

            chart
                .draw_series(LineSeries::new(
                    (-50..=50).map(|x| x as f32 / 50.0).map(|x| (x, x * x)),
                    &RED,
                ))
                .unwrap()
                .label("y = x^2")
                .legend(|(x, y)| PathElement::new(vec![(x, y), (x + 20, y)], RED));

            chart
                .configure_series_labels()
                .background_style(WHITE.mix(0.8))
                .border_style(BLACK)
                .draw()
                .unwrap();

            root.present().unwrap();
        })
    }

    pub fn format_bitmap_as_png(bitmap: Vec<u8>, width: u32, height: u32) -> Vec<u8> {
        let image =
            image::ImageBuffer::<image::Rgb<u8>, Vec<u8>>::from_raw(width, height, bitmap).unwrap();
        let mut buffer = BufWriter::new(Cursor::new(Vec::new()));
        image.write_to(&mut buffer, ImageFormat::Png).unwrap();
        let bytes: Vec<u8> = buffer.into_inner().unwrap().into_inner();
        bytes
    }

    pub fn plot(
        width: u32,
        height: u32,
        plotting_fun: BitMapPlotFun,
        format: BitMapOutputFormat,
    ) -> impl IntoResponse {
        let mut image_buffer = vec![0; (width as usize) * (height as usize) * 3];

        {
            let root =
                // BitMapBackend::with_buffer(&mut image_string, (width, height)).into_drawing_area();
                BitMapBackend::with_buffer(image_buffer.as_mut_slice(), (width, height)).into_drawing_area();
            plotting_fun(root);
        }

        let image = match format {
            BitMapOutputFormat::PNG => Self::format_bitmap_as_png(image_buffer, width, height),
        };

        let headers = [
            (
                header::CONTENT_TYPE,
                match format {
                    BitMapOutputFormat::PNG => "image/png",
                },
            ),
            (header::CONTENT_DISPOSITION, "inline"),
        ];
        (headers, image).into_response()
    }
}
